import sys
import json
from spout.parser import Parser
from spout.interpreter import Interpreter
from spout.spout import Runtime

def main():
    if len(sys.argv) >= 2:
        filename = sys.argv[1]
        data = ""
        with open(filename, 'r') as f:
            data = f.read()
        if len(sys.argv) >= 3 and sys.argv[2] == "--ast":
            ast = Runtime().AST(data)
            print(json.dumps(ast, indent=2))
        else:
            Runtime().run(data)
            
    else:
        parser = Parser()
        interpreter = Interpreter()
        print(f"\033[36mSpout\033[0m REPL v0.2")
        src = ""
        while True:
            data = input(">>> ") + "\n"
            if data == "exit\n":
                break
            elif data[-2:] == ";\n":
                src += data[:-1]
            else:
                src += data
                ast = parser.createAST(src)
                out = interpreter.eval_program(ast)
                src = ""
                if out == "undefined": print(f"\033[90mundefined\033[0m ") # 90m  is grey text
                else: 
                    if not isinstance(out, dict): print(out)
                    else: print(out["value"]) 
                
        

           


if __name__ == "__main__":
    main()



from .lexer import TokenType, Lexer, parseInt
from .environment import defineUndefined

class Parser:
    # have a private token list
    def __init__(self):
        self.tokens = []

    def notEOF(self):
        return self.tokens[0]["type"] != TokenType.EOF
    
    def consume(self):
        popped = self.tokens.pop(0)
        return popped
    
    def terminate_newline(self):
        while self.peek()["type"]==TokenType.Newline:
            self.consume()

    def peek(self):
        return self.tokens[0]

    def expect(self, type):
        if self.peek()["type"] == type:
            return self.consume()
        else:
            raise Exception("Expected " + type + " but got " + self.peek()["type"][0])

    def createAST(self, sourceCode):
        self.tokens = Lexer(sourceCode)
        ast = {
            "type": "Program",
            "body": []
        }
        while(self.notEOF()):
            ast["body"].append(self.parse_stmt())
        
        return ast

    def parse_stmt(self):
        ttype = self.peek()["type"]
        if ttype == TokenType.Let or ttype == TokenType.Const:
            return self.parse_var_declaration()
        elif ttype == TokenType.Fn:
            return self.parse_fn_declaration()
        elif ttype == TokenType.Whale:
            return self.parse_whale()
        return self.parse_expr()

    def parse_var_declaration(self):
        isConst = self.consume()["type"] == TokenType.Const
        name = self.expect(TokenType.Identifier)["value"]
        if self.peek()["type"] == TokenType.Newline:
            self.consume()
            if isConst:
                raise Exception("Constant values must have value at declaration")
            return {
                "type" : "VariableDeclaration",
                "name" : name,
                "const": False,
                "value": defineUndefined()
            }
        self.expect(TokenType.Equals)
        value = self.parse_expr()
        dec =  {
            "type": "VariableDeclaration",
            "name": name,
            "value": value,
            "const": isConst
        }
        self.expect(TokenType.Newline)
        return dec

    def parse_fn_declaration(self):
        self.consume()
        name = self.expect(TokenType.Identifier)["value"]
        arg = []
        if self.peek()["type"] == TokenType.OpenParen:
            self.consume()
            self.terminate_newline()
            while self.peek()["type"] != TokenType.CloseParen:
                arg.append(self.parse_expr())
                if self.peek()["type"] == TokenType.Comma:
                    self.consume()
                self.terminate_newline()
            self.expect(TokenType.CloseParen)
        else:
            while self.peek()["type"] != TokenType.Colon:
                arg.append(self.parse_expr())
                if self.peek()["type"] == TokenType.Comma:
                    self.consume()
        self.expect(TokenType.Colon)
        # Check if everything is an Identifier
        params = []
        for i in arg:
            if i["type"] != "Identifier":
                raise Exception("Function parameters must be identifiers")
            params.append(i["value"])

        body = []
        self.terminate_newline()
        if self.peek()["type"] == TokenType.OpenParen:
            self.consume()
            while(self.peek()["type"] != TokenType.EOF and self.peek()["type"] != TokenType.CloseParen):
                body.append(self.parse_stmt())
            self.expect(TokenType.CloseParen)
        else: 
            while(self.peek()["type"] != TokenType.EOF and self.peek()["type"] != TokenType.Newline):
                body.append(self.parse_stmt())
            
        return {
            "type": "FunctionDeclaration",
            "name": name,
            "parameters": params,
            "body": body
        }
        


    def parse_whale(self):
        self.consume()
        return {
            "type": "WhalingExpression",
            "value": self.parse_stmt()
            }

    def parse_expr(self):
        return self.parse_assignment_expr()
    
    def parse_assignment_expr(self):
        left = self.parse_bitwise_expr()
        if self.peek()["type"] == TokenType.Equals:
            self.consume()
            right = self.parse_bitwise_expr()
            return {
                "type": "AssignmentExpression",
                "left": left,
                "right": right
            }
        return left

    def parse_bitwise_expr(self):
        left = self.parse_shift_expr()
        while self.peek()["value"] == "&" or self.peek()["value"] == "|" or self.peek()["value"] == "^":
            op = self.consume()
            right = self.parse_shift_expr()
            left = {
                "type": "BinaryExpression",
                "operator": op["value"],
                "left": left,
                "right": right
            }
        return left

    def parse_shift_expr(self):
        left = self.parse_additive_expr()
        while self.peek()["value"] == "<<" or self.peek()["value"] == ">>":
            op = self.consume()
            right = self.parse_additive_expr()
            left = {
                "type": "BinaryExpression",
                "operator": op["value"],
                "left": left,
                "right": right
            }
        return left

    def parse_additive_expr(self):
        left = self.parse_multiplicative_expr()
        while self.peek()["value"] == "+" or self.peek()["value"] == "-":
            op = self.consume()
            right = self.parse_multiplicative_expr()
            left = {
                "type": "BinaryExpression",
                "operator": op["value"],
                "left": left,
                "right": right
            }
        return left
    
    def parse_multiplicative_expr(self):
        left = self.parse_exponential_expr()
        while self.peek()["value"] == "*" or self.peek()["value"] == "/" or self.peek()["value"] == "%":
            op = self.consume()
            right = self.parse_exponential_expr()
            left = {
                "type": "BinaryExpression",
                "operator": op["value"],
                "left": left,
                "right": right
            }
        return left
    
    def parse_exponential_expr(self):
        left = self.parse_call_member_expr()
        while self.peek()["value"] == "**":
            op = self.consume()
            right = self.parse_call_member_expr()
            left = {
                "type": "BinaryExpression",
                "operator": op["value"],
                "left": left,
                "right": right
            }
        return left

    def parse_call_member_expr(self):
        caller = self.parse_primary_expr()
        if self.peek()["type"] == TokenType.OpenParen:
            return self.parse_call_expr(caller)
        return caller

    def parse_call_expr(self, callee):
        call_expr = {
            "type": "CallExpression",
            "callee": callee,
            "args": self.parse_args()
        }
        if self.peek()["type"] == TokenType.OpenParen:
            call_expr = self.parse_call_expr(call_expr)
        return call_expr

    def parse_args(self):
        self.expect(TokenType.OpenParen)
        args = []
        while self.peek()["type"] != TokenType.CloseParen:
            args.append(self.parse_expr())
            if self.peek()["type"] == TokenType.Comma:
                self.consume()
        self.expect(TokenType.CloseParen)
        return args

    def parse_primary_expr(self):
        curr = self.peek()["type"]
        if curr == TokenType.Number:
            num = parseInt(self.consume()["value"])
            return {
                "type": "NumericLiteral",
                "value": num["value"],
                "base": num["base"],
                "dtype": "int"
            }
        elif curr == TokenType.String:
            return {
                "type": "StringLiteral",
                "value": self.consume()["value"],
                "dtype": "string"
            }
        elif curr == TokenType.Identifier:
            return {
                "type": "Identifier",
                "value": self.consume()["value"]
            }
        elif curr == TokenType.OpenParen:
            self.consume()
            expr = self.parse_expr()
            self.expect(TokenType.CloseParen)
            return expr
        elif curr == TokenType.Newline:
            self.consume()
            return
        elif curr == TokenType.EOF:
            print("what?")
        else:
            raise Exception("Unexpected token " + self.peek()["value"])
        





        



from enum import Enum, auto

class TokenType(Enum):
    Number = auto()
    Char = auto()
    Identifier = auto()
    String = auto()
    Let = auto()
    Const = auto()
    Fn = auto()
    Whale = auto()
    BinaryOperator = auto()
    UnaryOperator = auto()
    Equals = auto()
    Comma = auto()
    Colon = auto()
    OpenParen = auto()
    CloseParen = auto()
    Newline = auto()
    EOF = auto()

KEYWORDS = {
    "let": TokenType.Let,
    "const": TokenType.Const,
    "whale": TokenType.Whale,
    "spout": TokenType.Whale,
    "fn": TokenType.Fn
}

Logical = {
    "and": "&",
    "or": "|",
    "xor": "^",
    "lshift": ">>",
    "rshift": "<<"
}

BinOps = ["+", "-", "*", "/", "%", "&", "|", "^", "<<", ">>", "**", ","]
UnaOps = ["~"]



def Lexer(data):
    lexical = []
    #automatic newline insertion
    string = data.replace(";", "\n")
    string = string  + "\n\n"
    symbols = ["=", "+", "-", "*", "/", "%", "(", ")", "'", "&", "|", "^", "~", "!", "<", ">", "\n",",", ":", "<<", ">>", "**"]

    lexed = []
    temp = ""
    skipline = False
    
    i = 0
    while (i < len(string)):
        char = string[i]
        if skipline and char == "\n":
            skipline = False
        if skipline:
            i+=1
            continue
        if char != " " and char != "\t":
            temp += char
        if (i+1 < len(string)):
            if temp == "/" and string[i+1] == "/":
                lexed.append("<comment>")
                temp = ""
                skipline = True
            # handle strings
            if char == "\"":
                while(string[i+1] != "\""):
                    i+=1
                    temp += string[i]
                    if i >= len(string):
                        raise f"Unmatched double quotes (\") at {i}"
                temp+="\""
                i+=1
                lexed.append(temp)
                temp=""
                if i+1 == len(string):
                    continue
            if char == "'":
                while(string[i+1] != "'"):
                    i+=1
                    temp += string[i]
                    if i >= len(string):
                        raise f"Unmatched single quotes (\') at {i}"
                temp+="'"
                i+=1
                lexed.append(temp)
                temp=""
                if i+1 == len(string):
                    continue
            
            # Handle double symbols
            if char == "<" and string[i+1] == "<":
                temp += string[i+1]
                i+=1
            if char == ">" and string[i+1] == ">":
                temp += string[i+1]
                i+=1
            if char == "*" and string[i+1] == "*":
                temp += string[i+1]
                i+=1
            
            
            if string[i+1] == " " or string[i+1] == "\t" or string[i+1] in symbols or temp in symbols:
                if temp == "\n":
                    temp = "<newline>"
                if temp != "":
                    lexed.append(temp)  
                temp = ""
        i+=1
        
        # if temp != "":
        #     if temp == "\n": lexed.append("<newline>")
        #     else: lexed.append(temp)
    # print(lexed)
    tokens = tokenize(lexed)
    return tokens

def parseInt(value):
    last = ""
    base = 10
    if len(value)>2 and value[0] == "0" and value[1].isalpha():
        last = value[2:]
        if value[1] == 'b':
            base = 2
        elif value[1] == 'o':
            base = 8
        elif value[1] == 'x':
            base = 16
            last = str(int(last, base))
    else:
        last = value
    
    if last.isnumeric():
        return {"value": int(last, base), "base": base}
    return False
    

def token(value, type):
    return {"value": value, "type": type}

def tokenize(lex):
    tokens = []

    while(len(lex)):
        if lex[0] == "(":
            tokens.append(token(lex.pop(0),TokenType.OpenParen))
        elif lex[0] == ")":
            tokens.append(token(lex.pop(0),TokenType.CloseParen))
        elif lex[0] == ",":
            tokens.append(token(lex.pop(0),TokenType.Comma))
        elif lex[0] == ":":
            tokens.append(token(lex.pop(0),TokenType.Colon))
        elif lex[0] in BinOps:
            tokens.append(token(lex.pop(0),TokenType.BinaryOperator))
        elif lex[0] == "=":
            tokens.append(token(lex.pop(0),TokenType.Equals))
        elif lex[0] == "<comment>":
            lex.pop(0)
        elif lex[0] == "<newline>":
            tokens.append(token(lex.pop(0),TokenType.Newline))
        else:
            if parseInt(lex[0]):
                tokens.append(token(lex.pop(0), TokenType.Number))

            else:
                reserved = lex[0] in KEYWORDS
                if reserved:
                    tokens.append(token(lex[0], KEYWORDS[lex.pop(0)]))
                elif lex[0][0] == "\"":
                    tokens.append(token(lex.pop(0)[1:-1], TokenType.String))
                elif lex[0][0] == "'":
                    tokens.append(token(lex.pop(0), TokenType.Char))
                elif lex[0] in Logical:
                    tokens.append(token(Logical[lex.pop(0)], TokenType.BinaryOperator))
                else:
                    tokens.append(token(lex.pop(0), TokenType.Identifier))
            

    tokens.append(token("EndOfFile", TokenType.EOF))
    return tokens
            


        













from .parser import Parser
from .environment import defineUndefined, Environment
from .whale import render

class Interpreter:
    def __init__(self, env = Environment()):
        self.env = env
    def eval_program(self, ast, env=None):
        if env == None: env = self.env
        latest = defineUndefined()
        for stmt in ast["body"]:
            if stmt:
                latest = self.eval_stmt(stmt, env, False)
        # print(latest, not isinstance(latest, dict))
        # if not isinstance(latest, dict): return latest
        # return latest["value"]
        return latest
    
    def eval_stmt(self, stmt, env, whale):
        if stmt["type"] == "NumericLiteral":
            return { "type": "number", "value": stmt["value"], "dtype": stmt["dtype"]}
        elif stmt["type"] == "StringLiteral":
            return { "type": "string", "value": stmt["value"]}
        elif stmt["type"] == "Identifier":
            value = self.eval_identifier(stmt, env)
        elif stmt["type"] == "CallExpression":
            value = self.eval_call_expr(stmt, env, whale)
        elif stmt["type"] == "BinaryExpression":
            value = self.eval_binary_expr(stmt,env, whale)
        elif stmt["type"] == "AssignmentExpression":
            name = stmt["left"]["value"]
            value = self.eval_stmt(stmt["right"],env, whale)
            env.assign(name, value)
        elif stmt["type"] == "VariableDeclaration":
            name = stmt["name"]
            value = self.eval_stmt(stmt["value"],env, whale)
            isConst = stmt["const"]
            env.declare(name, value, isConst)
        elif stmt["type"] == "FunctionDeclaration":
            value = self.eval_function_declaration(stmt,env)
        elif stmt["type"] == "WhalingExpression":
            value = self.eval_stmt(stmt["value"],env, True)
        elif stmt["type"] == "Program":
            value = self.eval_program(stmt,env)
        else:
            raise Exception(f"Unknown statement type: {stmt['type']}\n{stmt}")
        return value
    
    def eval_identifier(self, stmt, env):
        return env.retrive(stmt["value"])
    
    def eval_function_declaration(self, stmt, env):
        fn = {
            "type": "function",
            "name": stmt["name"],
            "parameters": stmt["parameters"],
            "body": stmt["body"],
            "declarationEnv": env,
            "value": "undefined"
        }
        return env.declare(stmt["name"], fn, True)

    def eval_call_expr(self, stmt, env, whale):
        args = list(map(lambda arg: self.eval_stmt(arg, env, whale), stmt["args"]))
        fun = self.eval_stmt(stmt["callee"], env, whale)
        if fun["type"] == "NativeFunction":
            return fun["call"](args, env) or defineUndefined()

        elif fun["type"] == "function":
            if len(args) != len(fun["parameters"]):
                raise Exception(f"Expected {len(fun['parameters'])} arguments but got {len(args)}: in function {fun['name']}")
            callEnv = Environment(fun["declarationEnv"])
            for i in range(len(fun["parameters"])):
                callEnv.declare(fun["parameters"][i], args[i], True)
            return self.eval_program({"type": "program","body":fun["body"]}, callEnv) or defineUndefined()
        raise Exception(f"Cannot call non-function type: {fun['type']}")

    def eval_binary_expr(self, expr, env, whale):
        left = self.eval_stmt(expr["left"],env, whale)
        right = self.eval_stmt(expr["right"],env, whale)
        # check if both are numbers
        if left["type"] == "number" and right["type"] == "number":
            return self.eval_binary_expr_int(expr["operator"], left, right, whale)
        elif left["type"] == "string" or right["type"] == "string":
            return self.eval_binary_expr_str(expr["operator"], left, right)
        return defineUndefined()
    
    def eval_binary_expr_int(self, op, left_v, right_v, whale):
        value = defineUndefined()
        left = left_v["value"]
        right = right_v["value"]
        dtype = left_v["dtype"]
        if op == "+":
            value = left + right
        elif op == "-":
            value = left - right
        elif op == "*":
            value = left * right
        elif op == "/":
            value = left // right
        elif op == "%":
            value = left % right
        elif op == "**":
            value = left ** right
        elif op == "<<":
            value = left << right
        elif op == ">>":
            value = left >> right
        elif op == "&":
            value = left & right
        elif op == "|":
            value = left | right
        elif op == "^":
            value = left ^ right
        else:
            raise Exception(f"Unknown operator: {op}")
        
        if whale:
            render(op, value, left, right)
        return { "type": "number", "value": value, "dtype": dtype}
    
    
    def eval_binary_expr_str(self, op, left_v, right_v):
        value = ""
        left = left_v["value"]
        right = right_v["value"]
        ogRType = right_v["type"]

        if type(left) == int:
            left = str(left)
        if type(right) == int:
            right = str(right)
        if op == "+":
            value = left + right
        elif op == "-":
            value = left.replace(right, "")
        elif op == "*":
            if ogRType == "number":
                value = left * int(right)
            else:
                raise Exception(f"Cannot multiply string by string")
        elif op == "/":
            raise Exception(f"Cannot divide string by string")
        return { "type": "string", "value": value}




class Environment:
    def __init__(self, env=None):
        self.parent = env
        self.vars = {}
        self.consts = {}
        self.declare("print", defineNativeFunction(lambda args,scope: print("".join([str(item["value"]) for item in args]))), False)
        if self.parent == None:
            self.define_consts()

    def define_consts(self):
        self.declare("true", defineNum("int", 1), False)
        self.declare("false", defineNum("int", 0), False)
        self.declare("undefined", defineUndefined(), True)
        self.declare("__neg__", defineNum("int", -1), True )
    
    def declare(self, name, value, constant):
        if name in self.vars:
            raise Exception(f"Variable {name} already declared")
        self.vars[name] = value
        if constant: self.consts[name] = constant
        return value
    
    def assign(self, name, value):
        varEnv = self.resolve(name)
        if name in varEnv.consts:
            raise Exception(f"Cannot assign value to constant {name}")
        varEnv.vars[name] = value
        return value
    
    def retrive(self, name):
        varEnv = self.resolve(name)
        return varEnv.vars[name]

    def resolve(self, name):
        if name in self.vars:
            return self
        if self.parent != None:
            return self.parent.resolve(name)
        raise Exception(f"Variable {name} does not exist")

def defineNum(dtype, value):
    return { "type": "number", "value": value, "dtype": dtype}
def defineStr(value):
    return { "type": "string", "value": value}
def defineUndefined():
    return {"type": "undefined", "value": "undefined"}
def defineNativeFunction(call):
    return {"type": "NativeFunction", "call": call}





from .parser import Parser
from .interpreter import Interpreter

class Runtime:
    def __init__(self):
        self.version = "0.2"

    def run(self, code):
        ast = self.AST(code)
        interpreter = Interpreter()
        result = interpreter.eval_program(ast)
        return result or "undefined"
    
    def AST(self, code):
        parser = Parser()
        return parser.createAST(code)



OPS = {
    "+": "ADD",
    "-": "SUB",
    "*": "MULT",
    "/": "DIV",
    "**": "POW",
    "&": "AND",
    "|": "OR",
    "^": "XOR",
    "<<": "LSHIFT",
    ">>": "RSHIFT",
    "%": "MOD"
}
KWORDS ={
    "ADD": "+",
    "SUB": "-",
    "MULT": "*",
    "DIV": "/",
    "POW": "**",
    "AND": "&",
    "OR": "|",
    "XOR": "^",
    "LSHIFT": "<<",
    "RSHIFT": ">>",
    "MOD": "%"
}

def render(operator: str, result: int, var1: int, var2: int ):
    char, word = "", ""
    if operator in OPS:
        char = operator
        word = OPS[operator]
    else:
        word = operator.upper()
        char = KWORDS[word]
    operation, v_len= operation_builder(char, word)

    length = max(result.bit_length(), var1.bit_length(), var2.bit_length())
    form = "{0:0" + f"{length}"+ "b}"
    pad = " "* v_len
    result = pad + form.format(result) + f"={result}"
    var1 = pad + form.format(var1) + f"={var1}"
    var2 = operation + form.format(var2) + f"={var2}"
    strike = "-"* len(var1)
    print(var1, var2, strike, result, "\n", sep="\n")




def operation_builder(char, word):
    # visual_length= len(f"{word} {char} ")
    visual_length = 10
    op_str = f"\033[36m{word} {char}\033[0m " + " " * (visual_length - len(f"{word} {char} "))
    return (op_str,visual_length)




from setuptools import setup, find_packages

# Build command: python setup.py sdist
# Upload command: twine upload dist/*

setup(
    name = 'spoutlang',
    version = '0.2.2',
    author = 'Tasfiqul Tapu',
    author_email = 'iamtasfiqultapu@gmail.com',
    license = 'MIT',
    description = 'An interpreted programming language aimed at demistifying bit manipulation.',
    url = 'https://github.com/TasfiqulTapu/spout',
    py_modules = ['main'],
    package_dir={
        "spout": "spout"
    },
    install_requires = [],
    python_requires='>=3.7',
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    entry_points = '''
        [console_scripts]
        spout=main:main
    '''
)